/*
--------------------------------------------------------------------------------

 lcssr - longest common suffix (LCS) symbol ranking (SR) file (de-)compression

 (c) 2007 by Frank Schwellinger, frank.schwellinger@gmx.de

 compiled with: 

 g++ -O2 -march=pentiumpro -Wall -fomit-frame-pointer -s -o lcssr lcssr-0.2.cpp

--------------------------------------------------------------------------------

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version. This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
details at http://www.gnu.org/copyleft/gpl.html.

--------------------------------------------------------------------------------
Program description
--------------------------------------------------------------------------------
lcssr is a derivative of symbra-0.2 available at

http://de.geocities.com/ocamyd/symbra-0.2.cpp.gz

Both progrmas are symbol ranking file (de-)compressors somewhat similar
to Peter Fenwicks symbol ranking algorithm 'srank', available at:

ftp://ftp.cs.auckland.ac.nz/pub/staff/peter-f/srank.c

The srank algorithm, designed for speed rather than good compression, is a
4-way set associative cache with a LRU update in order-3 contexts. When a
symbol is encountered it is encoded with 1, 3, or 4 bits according to its
rank in a queue of length 3, then moved to the front. Long runs of rank-0
are run length encoded using 12 bits to encode the length of the run. A miss
is coded using pseudo-MTF (swapping the symbol with another symbol about half
way to the front) in an order-0 context using 7 bits for the first 32 symbols
and 12 bits for the rest.

In contrast to the recently released C++ versions sr2 by Matt Mahoney and sr3 by
Francesco Antonio Nania (see http://cs.fit.edu/~mmahoney/compression/text.html),
which are modernized and improved versions of srank, lcssr follows more the
classic approach of srank: e.g. in that it uses no arithmetic coding. While
being similar to srank, lcssr still differs substantially from srank in the
way it ranks symbols, the generation of codes, the handling of the output and
in the way it handles the contexts.

In lcssr the context queues are removed and substituted with a suffix match
search. A single queue of 3 bytes length remains, which holds the last 3 results
of the search for the longest common suffix (LCS). In a first step lcssr makes
internal static code assignments similar to srank: an encountered symbol is
assigned to a code of 1, 2, or 3 bits length according to its queue position.
Runs of 9 rank-0s are assigned to a 5 bit RLE code. A miss gets assigned to an
8 or 13 bit code determined by the symbols MTF-index in an order-1 context,
where the context is the symbol following the longest common suffix, then the
symbol is moved (not swapped) about the half way to the front.

--------------------------------------------------------------------------------
Static code assignments
--------------------------------------------------------------------------------
Code            Length    Description
--------------  --------  ------------------------------------------------------
0                 1 bit   rank0: current symbol matches symbol following LCS
10                2 bits  rank1: current symbol matches symbol following 2nd LCS
110               3 bits  rank2: current symbol matches symbol following 3rd LCS
1110 xxxx       4+4 bits  current symbols order-1 MTF-index <16
11110 xxxxxxxx  5+8 bits  current symbols order-1 MTF-index >=16
11110 00000000  5+8 bits  EOB indicator
11111             5 bits  Run of 9 rank-0s
--------------------------------------------------------------------------------

But these are only internal assignments, in a second step lcssr uses adaptive
codes for the final encoding. That means the above listed code assignments are
encoded by their adaptively updated frequencies in an order-2 context, where the
context is build from the previous two static codes packed into one byte. The most
frequent static code seen in an order-2 context now gets encoded with the
shortest code and the least frequent static code seen in an order-2 context gets
encoded with the longest code.

-----------------------------------------------------
Adaptive code assignments
-----------------------------------------------------
Code   Length   Description
-----  -------  -------------------------------------
0       1 bit   static-code order-2-frequency rank-0
10      2 bits  static-code order-2-frequency rank-1
110     3 bits  static-code order-2-frequency rank-2
1110    4 bits  static-code order-2-frequency rank-3
11110   5 bits  static-code order-2-frequency rank-4
11111   5 bits  static-code order-2-frequency rank-5

(eventually followed by a 4 or 8 bit order-1 
MTF-index, or an 8 bit EOB marker)
-----------------------------------------------------

lcssr pushes the generated codes and MTF-indices and EOB marks into a 64 bit
FIFO queue, from where they are bytewise pulled out into a blockbuffer, while a
second blockbuffer gets filled with the original uncompressed bytes. lcssr
manages the output in small blocks of variable length to reduce the expansion
of data that e.g. srank produces when it processes precompressed data. If the
compression of a block fails, the blockbuffer with the original uncompressed
data is used for that block instead, thus exchanging sranks expansion
(<= ~40.9%) with a much smaller one (<= ~1.3%), which is introduced by the block
structures and by specific codes needed for the block management: e.g the first
byte of a block, which is used as the block header indicating the type (1 bit)
and the length (7 bits) of a block, and the (adpative) EOB marker (<= 13 bits)
which indicates the end of a block. The last block is always transmitted
uncompressed with a 3 byte block header.

--------------------------------------------------------------------------
Block Structures
--------------------------------------------------------------------------
Bit 1  Type          Bit 2-8              followed by
-----  ------------  -------------------  --------------------------------
1      compressed    7 bits block length  block data + EOB
0      uncompressed  7 bits block length  block data
0      uncompressed  7 zeroes=last block  16 bit block length + block data
--------------------------------------------------------------------------

The LCS (longest common suffix) search is the center of the algorithm. It is
choosen for its flexibility in finding matching contexts of variable lengths.
The LCS search is implemented as an unbalanced binary search tree, which seems
to be too slow for practical use, but it allows searches in buffers up to a
size of 455 MB, assumed the appropriate free memory (buffer size x 9) is
available. For this proof-of-concept implementation the binary search tree
solution may be good enough. The buffer size and the search depth (suffix match
limit) can be configured by commandline argument. Choosing a limit for the suffix
matches reduces the time that is spent for searches. In my tests a limit of 64
seems to be a good choice for most cases. Choosing a higher limit may improve
compression only by a small degree, but it can dramatically increase the time
used for searches. When a large buffer size is choosen, lcssr may find matching
sequences that are far apart from each other, but a larger buffer doesn't
necessarily mean better compression. Compared to LZ algorithms lcssr needs
larger buffer sizes to achieve comparable results.

What lcssr needs is a more efficient search for the longest common suffix,
although this implementation is already the fastet solution from a few approaches
i have tested. Maybe someone can make it better and if you are interested in the
topic you may want to take a look at Yuta Mori's web page at:
http://homepage3.nifty.com/wpage/links/suffix_arrays.html,
where you can find some links to related source codes and papers dealing with
suffix arrays and suffix sorting.

-----------------------------------------------------------------------------
File Header Structure
-----------------------------------------------------------------------------
Bytes  Description
-----  ----------------------------------------------------------------------
    1  Magic Byte = 234
    1  Program Version
    1  Program Subversion
    1  Buffermode (0-9)
    1  Limitmode (0-9)
-----------------------------------------------------------------------------

Beside the memory used for the binary search tree (buffer size x 9), lcssr
allocates only < 128 kbytes for the handling of the symbol ranking, the
generation of codes and the block management.

------------------------------------------------------------------------
Memory Usage
------------------------------------------------------------------------
Usage      Description
-------  ---------------------------------------------------------------
9MB-4GB  binary search tree for suffix match searches: buffer length x 9
64.0 KB  order-1 MTF-table (context is symbol following LCS)
 7.5 KB  order-2 code frequencies (context is last 2 codes packed in 1 Byte)
17.9 KB  2 block buffers (holding compressed/uncompressed data)
xx.x KB  other variables and program code
------------------------------------------------------------------------

Conclusion: The algorithm described above has some similarities to LZ77 or
LZP, and it can achieve compression results almost comparable to those of the
LZ77 algorithms (and sometimes even surpass them), and it is also similar
to symbol ranking algorithms like srank/sr2/sr3, but it doesn't fit in either
category. It may be best described as a hybrid LZ/SR algorithm. Despite the slow
LCS search it is a compact algorithm for data compression.

Note: Only input files with a file size up to 4 GB are supported by this version
(this is not tested by the program).

--------------------------------------------------------------------------------
HISTORY
--------------------------------------------------------------------------------
version  from date   until date   changes
-------  ----------- -----------  ----------------------------------------------
0.1      01-Dec-2007 01-Dec-2007  initial version
0.2      02-Dec-2007 03-Dec-2007  exclusion of symbol following LCS from 4 bit
                                  mtf zone; program description edited
--------------------------------------------------------------------------------
*/
//------------------------------------------------------------------------------
// INCLUDED HEADER FILES
//------------------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
//------------------------------------------------------------------------------
// OTHER COMPILER DIRECTIVES
//------------------------------------------------------------------------------
//#define VERBOSEMODE 1 //uncomment for testing
#define UINT8 unsigned char //unsigned 8bit integer
#define UINT32 unsigned int //unsigned 32bit integer
#define SINT32 int //signed 32bit integer
#define UINT64 unsigned long long int //unsigned 64bit integer
//------------------------------------------------------------------------------
// GLOBALS
//------------------------------------------------------------------------------
//---Program Version
const UINT8 VERSION=0;
const UINT8 SUBVERSION=2;
//---Program modes
UINT8 buffermode;
UINT8 limitmode;
//---Generated file header
UINT32 headerlength;
const UINT8 MAGICBYTE=234; //identifies generated stream
//---Suffix match queue
UINT32 rank0=0,rank1=0,rank2=0;
//---Move-to-front table
UINT8 mtf[256][256];
UINT32 mtfpos=0;
UINT32 mtfctx=0;
UINT8 mtftemp=0;
UINT32 mtftarget=0;
//---Number of code assignments
const UINT32 NCODES=6;
//---Code Frequencies
typedef struct co
{
	UINT8 code;
	UINT32 count;
}codetype;
codetype codefreq[256][NCODES];
UINT32 cfctx=0;
//---64bit FIFO queue
SINT32 bits=0;
UINT64 q=0;
const SINT32 QLEN=(sizeof(q)*8);
const UINT64 QMASK=0xFFFFFFFFFFFFFFFFULL;
const UINT32 RUNLENGTHLIMIT=9; //above this value FIFO queue may overflow
//---Block buffers
const SINT32 BLOCKLIMIT=127;
UINT8 outbuf[(BLOCKLIMIT*8*RUNLENGTHLIMIT)+13];
UINT8 orgbuf[(BLOCKLIMIT*8*RUNLENGTHLIMIT)+13];
UINT32 orgbufptr=0;
//---other globals used by encoder and decoder
UINT32 ch=0;
UINT32 outbytes=0, inbytes=0;
double totaltime;
FILE *in,*out;
//---Suffix match search
UINT32 matchlen=0;
UINT32 inbufsize=0; 
UINT8 *inbuf; //input buffer (dynamic allocation at runtime)
UINT32 *bstrightchild; //binary search tree right child nodes (dynamic allocation at runtime)
UINT32 *bstleftchild; //binary search tree left child nodes (dynamic allocation at runtime)
UINT32 inbufprefix=257; //Size in bytes of input buffer prefix
UINT32 inbufpos=257; //(Start) position in the input buffer
UINT8 matchchar=0; //char following longest suffix match
UINT8 tmpmatchchar=0;
UINT32 bstnode=0;
UINT32 bstparent=0;
UINT32 bstcomparison=0;
UINT32 bstlimit=0;
//------------------------------------------------------------------------------
// RESET CODE FREQUENCIES
//------------------------------------------------------------------------------
inline void resetcodefreq()
{
	UINT32 i,ii;
	//---reset code frequencies
	for(ii=0;ii<256;ii++) {
		for(i=0;i<NCODES;i++) {
			codefreq[ii][i].code=(UINT8)i;
			codefreq[ii][i].count=0;
		}
	}
	cfctx=0;
}
//------------------------------------------------------------------------------
// INIT MEMORY
//------------------------------------------------------------------------------
inline void initmemory()
{
	UINT32 i,ii;
	//---Init MTF
	for(ii=0;ii<256;ii++) {
		for(i=0;i<256;i++) {
			mtf[ii][i]=i;
		}
	}

	//---Init code frequencies
	resetcodefreq();

	switch (buffermode) {
		case 0: inbufsize=1*1024*1024;break;
		case 1: inbufsize=2*1024*1024;break;
		case 2: inbufsize=4*1024*1024;break;
		case 3: inbufsize=8*1024*1024;break;
		case 4: inbufsize=16*1024*1024;break;
		case 5: inbufsize=32*1024*1024;break;
		case 6: inbufsize=64*1024*1024;break;
		case 7: inbufsize=128*1024*1024;break;
		case 8: inbufsize=256*1024*1024;break;
		case 9: inbufsize=455*1024*1024;break;
	}

	switch (limitmode) {
		case 0: bstlimit=0; break;
		case 1: bstlimit=1; break;
		case 2: bstlimit=2; break;
		case 3: bstlimit=4; break;
		case 4: bstlimit=8; break;
		case 5: bstlimit=16; break;
		case 6: bstlimit=32; break;
		case 7: bstlimit=64; break;
		case 8: bstlimit=128; break;
		case 9: bstlimit=256; break;
	}

#ifdef VERBOSEMODE
	printf("Buffer size: %d bytes (Memory usage: %d bytes)\n",inbufsize,inbufsize*9);
	printf("Suffix match limit: %d bytes\n",bstlimit);
#endif

	//---Init Input Buffer
	inbuf=(UINT8 *)malloc(sizeof(UINT8)*(inbufsize+inbufprefix));
	if (inbuf==NULL) {
		printf("Memory allocation failed\n");
		exit(1);
	}
	for(i=0;i<inbufsize+inbufprefix;i++) {
		inbuf[i]=i%256;
	}
	
	//---Init binary search tree
	bstrightchild=(UINT32 *)malloc(sizeof(UINT32)*(inbufsize+inbufprefix));
	if (bstrightchild==NULL) {
		printf("Memory allocation failed\n");
		exit(1);
	}
	bstleftchild=(UINT32 *)malloc(sizeof(UINT32)*(inbufsize+inbufprefix));
	if (bstleftchild==NULL) {
		printf("Memory allocation failed\n");
		exit(1);
	}
	for(i=0;i<inbufsize+inbufprefix;i++) {
		bstrightchild[i]=0;
		bstleftchild[i]=0;
	}

}
//------------------------------------------------------------------------------
// REBUILD BINARY SEARCH TREE
//------------------------------------------------------------------------------
inline void rebuildbst()
{
	UINT32 i;
#ifdef VERBOSEMODE
	printf("rebuilding binary search tree\n");
#endif
	//---reset binary search tree
	for(i=0;i<inbufsize+inbufprefix;i++) {
		bstrightchild[i]=0;
		bstleftchild[i]=0;
	}
	//---partially rebuild binary search tree
	for(i=(inbufsize/2)+inbufprefix;i<inbufprefix+inbufsize-1;i++) {
		//---update input buffer
		inbufpos++;
		inbuf[inbufpos]=inbuf[i];
		bstnode=inbufprefix;
		while(1) {
			if(bstnode==0) {
				bstcomparison=0;
				while(1) {
					if(inbuf[inbufpos-bstcomparison]>
					        inbuf[bstparent-bstcomparison]) {
						bstrightchild[bstparent]=inbufpos;
						break;
					}
					if(inbuf[inbufpos-bstcomparison]<
					        inbuf[bstparent-bstcomparison]) {
						bstleftchild[bstparent]=inbufpos;
						break;
					}
					bstcomparison++;
					if(bstcomparison>bstlimit) {
						break;
					}
				}
				break;
			} else {
				bstcomparison=0;
				while(1) {
					if(inbuf[inbufpos-bstcomparison]>
					        inbuf[bstnode-bstcomparison]) {
						bstparent=bstnode;
						bstnode=bstrightchild[bstnode];
						break;
					}
					if(inbuf[inbufpos-bstcomparison]<
					        inbuf[bstnode-bstcomparison]) {
						bstparent=bstnode;
						bstnode=bstleftchild[bstnode];
						break;
					}
					bstcomparison++;
					if(bstcomparison>bstlimit) {
						bstparent=bstnode;
						bstnode=0;
						break;
					}
				}
			}
		}
	}
}
//------------------------------------------------------------------------------
// UPDATE BINARY SEARCH TREE
//------------------------------------------------------------------------------
inline void updatebst()
{
	//---update input buffer
	inbufpos++;
	if(inbufpos==(UINT32)(inbufsize+inbufprefix)) {
		inbufpos=inbufprefix;
		rebuildbst();
	}
	inbuf[inbufpos]=ch;

	//---update binary search tree and search longest suffix-match
	matchlen=0;
	tmpmatchchar=0;
	matchchar=0;
	bstnode=inbufprefix;
	while(1) {
		if(bstnode==0) {
			bstcomparison=0;
			while(1) {
				if(inbuf[inbufpos-bstcomparison]>
					inbuf[bstparent-bstcomparison]) {
					bstrightchild[bstparent]=inbufpos;
					break;
				}
				if(inbuf[inbufpos-bstcomparison]<
					inbuf[bstparent-bstcomparison]) {
					bstleftchild[bstparent]=inbufpos;
					break;
				}
				bstcomparison++;
				if(bstcomparison>bstlimit) {
					break;
				}
			}
			break;
		} else {
			if(bstnode+1<(UINT32)(inbufprefix+inbufsize))
				tmpmatchchar=inbuf[bstnode+1];
			bstcomparison=0;
			while(1) {
				if(inbuf[inbufpos-bstcomparison]>
					inbuf[bstnode-bstcomparison]) {
					bstparent=bstnode;
					bstnode=bstrightchild[bstnode];
					break;
				}
				if(inbuf[inbufpos-bstcomparison]<
					inbuf[bstnode-bstcomparison]) {
					bstparent=bstnode;
					bstnode=bstleftchild[bstnode];
					break;
				}
				bstcomparison++;
				if(bstcomparison>=matchlen) {
					//---longest match
					matchlen=bstcomparison;
					matchchar=tmpmatchchar;
					//---store matchchar in an 3 byte mtf-table
					if(matchchar==rank1){
						rank1=rank0;
						rank0=matchchar;
					}else{
						if(matchchar==rank2){
							rank2=rank1;
							rank1=rank0;
							rank0=matchchar;
						}else{
							if(matchchar!=rank0){
								rank2=rank1;
								rank1=rank0;
								rank0=matchchar;
							}
						}
					}
				}
				if(bstcomparison>bstlimit) {
					bstparent=bstnode;
					bstnode=0;
					break;
				}
			}
		}
	}
	//---end of bst
}
//------------------------------------------------------------------------------
// WRITE HEADER
//------------------------------------------------------------------------------
inline void writeheader(FILE *out)
{
	fputc(MAGICBYTE,out);
	fputc(VERSION,out);
	fputc(SUBVERSION,out);
	fputc(buffermode,out);
	fputc(limitmode,out);
	headerlength=5;
}
//------------------------------------------------------------------------------
// READ HEADER
//------------------------------------------------------------------------------
inline SINT32 readheader(FILE *in)
{
	if(fgetc(in)!=MAGICBYTE)return 1;
	if(fgetc(in)!=VERSION)return 1;
	if(fgetc(in)!=SUBVERSION)return 1;
	buffermode=fgetc(in);if(buffermode>9)return 1;
	limitmode=fgetc(in);if(limitmode>9)return 1;
	headerlength=5;
	return 0;
}
//------------------------------------------------------------------------------
// UPDATE CODE FREQUENCIES
//------------------------------------------------------------------------------
inline void updatecodefreq(UINT32 code)
{
	UINT32 i=0;
	UINT8 tmpcode=0;
	UINT32 tmpcount=0;
	//---update code frequency and keep table sorted by frequencies
	for(i=0;i<NCODES;i++) {
		if(codefreq[cfctx][i].code==(UINT8)code){
			codefreq[cfctx][i].count++;
			while((i>0) && (codefreq[cfctx][i].count>codefreq[cfctx][i-1].count)){
				tmpcode=codefreq[cfctx][i].code;
				tmpcount=codefreq[cfctx][i].count;
				codefreq[cfctx][i].code=codefreq[cfctx][i-1].code;
				codefreq[cfctx][i].count=codefreq[cfctx][i-1].count;
				codefreq[cfctx][i-1].code=tmpcode;
				codefreq[cfctx][i-1].count=tmpcount;
				i--;
			}
			i=NCODES;
		}
	}
	cfctx=((cfctx<<4)+code)&0x000000FF;
}
//------------------------------------------------------------------------------
// ADD CODE TO 64 BIT QUEUE
//------------------------------------------------------------------------------
inline void addcode(UINT32 code)
{
	UINT32 i=0;
	//---search code frequency rank
	while(i<NCODES) {
		if(codefreq[cfctx][i].code==(UINT8)code){
			break;
		}
		i++;
	}
#ifdef VERBOSEMODE
	if(i>(NCODES-1))printf("error in getcode(): code not in range 0..%d\n",NCODES-1);
#endif
	//---add code to queue
	switch (i){
		case 0:
			//binary code 0
			bits++;
			q<<=1;
			break;
		case 1:
			//binary code 10
			bits+=2;
			q<<=2;
			q+=0x0000000000000002ULL;
			break;
		case 2:
			//binary code 110
			bits+=3;
			q<<=3;
			q+=0x0000000000000006ULL;
			break;
		case 3:
			//binary code 1110
			bits+=4;
			q<<=4;
			q+=0x000000000000000EULL;
			break;
		case 4:
			//binary code 11110
			bits+=5;
			q<<=5;
			q+=0x000000000000001EULL;
			break;
		case 5:
			//binary code 11111
			bits+=5;
			q<<=5;
			q+=0x000000000000001FULL;
			break;
	}
	updatecodefreq(code);
}
//------------------------------------------------------------------------------
// GET MTF
//------------------------------------------------------------------------------
inline void getmtf()
{
	UINT32 i;
	//---get mtf position
	for(i=0;i<256;i++) {
		if(mtf[mtfctx][i]==ch){
			mtfpos=i;
			i=256;
		}
	}
}
//------------------------------------------------------------------------------
// UPDATE MTF
//------------------------------------------------------------------------------
inline void updatemtf()
{
	UINT32 i;
	//---move to front
	if(mtfpos!=0){
		mtftarget=(UINT32)(mtfpos>>1);
		mtftemp=mtf[mtfctx][mtfpos];
		for(i=mtfpos;i>mtftarget;i--) {
			mtf[mtfctx][i]=mtf[mtfctx][i-1];
		}
		mtf[mtfctx][mtftarget]=mtftemp;
	}

	//---update MTF context
	mtfctx=matchchar;

	//---if matchchar is in mtf-4-bit zone move it out
	for(i=0;i<16;i++) {
		if(mtf[mtfctx][i]==matchchar){
			mtfpos=i;
			for(i=mtfpos;i<16;i++) {
				mtf[mtfctx][i]=mtf[mtfctx][i+1];
			}
			mtf[mtfctx][16]=matchchar;
			i=16;
		}
	}
}
//------------------------------------------------------------------------------
// ENCODE
//------------------------------------------------------------------------------
SINT32 encode(char *infile,char *outfile)
{
	clock_t begintime = clock();
	UINT32 i=0;
	UINT32 endoffile=(UINT32)EOF;
	UINT32 outbufptr=0;
	UINT32 runlengthcounter=0;
	const SINT32 BLOCKMINUS=2+(QLEN/8);
	const UINT32 BLOCKTRIGGER=(UINT32)(BLOCKLIMIT-BLOCKMINUS);

	//---init memory
	initmemory();

	//---open files
	in=fopen(infile,"rb");
	out=fopen(outfile,"wb");
	if(in==NULL || out==NULL) {
		fclose(in);
		fclose(out);
		printf("can't open files\n");
		return 1;
	}

	//---write header
	writeheader(out);

	//---process input file
	while(ch=(UINT32)fgetc(in),ch!=endoffile) {
		inbytes++;

		//---get mtf position
		getmtf();

		//---build codes
		if(rank0==ch) {
			runlengthcounter++;
			if(runlengthcounter==RUNLENGTHLIMIT){
				//RLE
				addcode(5);
				runlengthcounter=0;
			}
		} else {
			while(runlengthcounter>0){
				//Rank-0
				addcode(0);
				runlengthcounter--;
			}
			if(rank1==ch) {
				//Rank-1
				addcode(1);
			}else{
				if(rank2==ch){
					//Rank-2
					addcode(2);
				}else{
					if(mtfpos<16) {
						//MTF<16
						addcode(3);
						bits+=4;
						q<<=4;
						q+=((UINT64)mtfpos)&0x000000000000000FULL;
					}else{
						//MTF>=16
						addcode(4);
						bits+=8;
						q<<=8;
						q+=((UINT64)mtfpos)&0x00000000000000FFULL;
					}
				}
			}
		}

#ifdef VERBOSEMODE
		if(bits>QLEN){
			printf("error in encode(): queue overflow 1\n");
			return 1;
		}
#endif
		//---update orgbuf with original uncompressed values
		orgbufptr++;
		orgbuf[orgbufptr]=(UINT8)ch;
		//---pull queue bytewise to outbuf
		while(bits>=8){
			outbufptr++;
			if(bits>=8){
				bits-=8;
				outbuf[outbufptr]=(UINT8)((q>>bits)&0x00000000000000FFULL);
			}else{
				bits=0;
				outbuf[outbufptr]=(UINT8)(q&0x00000000000000FFULL);
			}
			if(bits>0){
				q=q&(QMASK>>(QLEN-bits));
			}else{
				q=0;
			}
			//---block?
			if(outbufptr==BLOCKTRIGGER){
				if((outbufptr+BLOCKMINUS)>=orgbufptr){
#ifdef VERBOSEMODE
					if(orgbufptr>(UINT32)BLOCKLIMIT){
						printf("error in encode(): allowed block buffer length exceeded 1\n");
					}
#endif
					//---uncompressed block
					orgbuf[0]=orgbufptr; //block id 0 + 7bits block length
					fwrite(orgbuf,sizeof(UINT8),orgbufptr+1,out);
					outbytes+=orgbufptr+1;

					//---Reset code frequencies
					//---(otherwise decoders code-frequencies are out of sync)
					resetcodefreq();
				}else{
					//---compressed block
					if(bits>(QLEN-13)){ //free space for EOB in queue?
						outbufptr++;
						bits-=8;
						outbuf[outbufptr]=(UINT8)((q>>bits)&0x00000000000000FFULL);
						q=q&(QMASK>>(QLEN-bits));
						outbufptr++;
						bits-=8;
						outbuf[outbufptr]=(UINT8)((q>>bits)&0x00000000000000FFULL);
						q=q&(QMASK>>(QLEN-bits));
					}
					//---add EOB (code 11110 + 8 zeroes) to queue
					addcode(4);
					bits+=8;
					q<<=8;
					//---add remaining queue to outbuf
#ifdef VERBOSEMODE
					if(bits>QLEN){
						printf("error in encode(): queue overflow 2\n");
						return 1;
					}
#endif
					while(bits>0){
						outbufptr++;
						if(bits>=8){
							bits-=8;
							outbuf[outbufptr]=(UINT8)((q>>bits)&0x00000000000000FFULL);
						}else{
							bits=0;
							outbuf[outbufptr]=(UINT8)(q&0x00000000000000FFULL);
						}
						if(bits>0){
							q=q&(QMASK>>(QLEN-bits));
						}else{
							q=0;
						}
					}
					//---write compressed block out
#ifdef VERBOSEMODE
					if(outbufptr>(UINT32)BLOCKLIMIT){
						printf("error in encode(): allowed block buffer length exceeded 2\n");
					}
#endif
					outbuf[0]=(0x80+(UINT8)outbufptr); //block id 1 + 7 bits block length
					fwrite(outbuf,sizeof(UINT8),outbufptr+1,out);
					outbytes+=outbufptr+1;
				}
				//---reset block buffers and queue
				bits=0;
				q=0;
				orgbufptr=0;
				outbufptr=0;
			}//of block
		}

		//---update bst
		updatebst();
		//---move to front and update MTF context
		updatemtf();

#ifdef VERBOSEMODE
		//---display result
		if(inbytes%0xFFFF==0)printf("%d bytes in %d bytes out ratio: %f\n",inbytes,outbytes,(float)outbytes/inbytes);
#endif
	}//eof

	//---write uncompressed rest out
	fputc(0x00,out); //last block indicator
	fputc((orgbufptr&0x0000FF00)>>8,out); //block length
	fputc(orgbufptr&0x000000FF,out); //block length
	for(i=0;i<orgbufptr;i++){
		fputc(orgbuf[i+1],out);
	} 
	outbytes+=orgbufptr+3;

	//---close Files
	fclose(in);
	fclose(out);

	//---add header length
	outbytes+=headerlength;

	//---display result
	printf ("%d Bytes encoded to %d Bytes\n",inbytes,outbytes);
	printf ("Ratio: %f = %f bpb\n",(float)outbytes/inbytes,((float)outbytes/inbytes)*8);
	totaltime=((double)(clock()-begintime))/CLOCKS_PER_SEC;
	printf("Time : %.3f sec\n",totaltime);
	printf("Speed: %.3f Mbyte/s in, %.3f Mbyte/s out\n",(inbytes/totaltime)/(1024*1024),(outbytes/totaltime)/(1024*1024));

	//---end of encoding
	return 0;
}
//------------------------------------------------------------------------------
// DECODE
//------------------------------------------------------------------------------
SINT32 decode(char *infile,char *outfile)
{
	clock_t begintime = clock();
	UINT8 stop=0,stopblock=0;
	UINT8 blockid=0;
	UINT8 blockmode=0;
	UINT32 blockidx=0;
	UINT32 b=0,run=0,r=0;
	SINT32 bctr=0;
	const SINT32 MAXCODELENGTH=(SINT32)(NCODES-1); 

	//---open files
	in=fopen(infile,"rb");
	out=fopen(outfile,"wb");
	if(in==NULL || out==NULL) {
		fclose(in);
		fclose(out);
		printf("can't open files\n");
		return 1;
	}

	//---read header
	if(readheader(in)==1){
		printf("unknown input file format\n");
		return 1;
	}

	//---init memory
	initmemory();

	stop=0;
	while(stop==0){
		blockid=fgetc(in);
		inbytes++;
		if((blockid&0x80)==0x00){
			if(blockid==0x00){
				//---last block
				blockmode=2;
				orgbufptr=(UINT32)((fgetc(in)*256)+fgetc(in)); //16bit block length
				inbytes+=2;
				fread(orgbuf,sizeof(UINT8),orgbufptr,in);
				inbytes+=orgbufptr;
				fwrite(orgbuf,sizeof(UINT8),orgbufptr,out);
				outbytes+=orgbufptr;
				stop=1;
			}else{
				//---uncompressed block
				blockmode=0;
				orgbufptr=(UINT32)(blockid&0x7F); //7bit block length
				fread(orgbuf,sizeof(UINT8),orgbufptr,in);
				inbytes+=orgbufptr;
			}
		}else{
			//---compressed block
			blockmode=1;
			orgbufptr=(UINT32)(blockid&0x7F); //7bit block length
			fread(orgbuf,sizeof(UINT8),orgbufptr,in);
			inbytes+=orgbufptr;
		}

		if(blockmode==0){ //---uncompressed block
			for(blockidx=0;blockidx<orgbufptr;blockidx++){
				ch=orgbuf[blockidx];
				fputc(ch,out);
				outbytes++;
				//---update bst
				updatebst();
				//---get mtf position
				getmtf();
				//---move to front and update MTF context
				updatemtf();
			}
			//---Reset code frequencies
			//---(Sync with encoder)
			resetcodefreq();
		}else{//---compressed or last block
			if(blockmode==1){//---compressed block
				q=0ULL;
				bits=0;
				blockidx=0;
				stopblock=0;
				while(stopblock==0){

					//---scan bitwise to reconstruct frequency-code	
					bctr=0;
					while(bctr<MAXCODELENGTH){
						if(bits<1){
							q<<=8;
							blockidx++;
							q+=(UINT64)orgbuf[blockidx-1];
							bits+=8;
						}
						if(bits>1){
							b=(UINT32)(q>>(bits-1));
							q=q&(QMASK>>((QLEN+1)-bits));
							bits-=1;
						}else{
							b=(UINT32)q;
							q=0ULL;
							bits-=1;
						}
						if(b==1){
							bctr++;
						}else{
							break;
						}
					}

					bctr=codefreq[cfctx][bctr].code; //decode frequency-code
					updatecodefreq(bctr); //keep code frequencies in sync with encoder
	
					run=1;
					switch (bctr){
						case 0: ch=rank0; break;
						case 1: ch=rank1; break;
						case 2: ch=rank2; break;
						case 3:
							//---read 4 bit MTF-index
							if(bits<4){
								q<<=8;
								blockidx++;
								q+=(UINT64)orgbuf[blockidx-1];
								bits+=8;
							}
							if(bits>4){
								b=(UINT32)(q>>(bits-4));
								q=q&(QMASK>>((QLEN+4)-bits));
								bits-=4;
							}else{
								b=(UINT32)q;
								q=0;
								bits-=4;
							}
							//---rank <16
							ch=mtf[mtfctx][b];
#ifdef VERBOSEMODE
							if(b>=16){
								printf("error in decode(): MTF index >= 16\n");
							}
#endif
							break;
						case 4:
							//---read 8 bit MTF-index or EOB
							if(bits<8){
								q<<=8;
								blockidx++;
								q+=(UINT64)orgbuf[blockidx-1];
								bits+=8;
							}
							if(bits>8){
								b=(UINT32)(q>>(bits-8));
								q=q&(QMASK>>((QLEN+8)-bits));
								bits-=8;
							}else{
								b=(UINT32)q;
								q=0ULL;
								bits-=8;
							}
							if(b!=0){
								//---rank >=16
								ch=mtf[mtfctx][b];
							}else{
								//---EOB
								stopblock=1;
								run=0;
							}
							break;
						case 5:
							//---run of rank-0s
							ch=rank0;
							run=RUNLENGTHLIMIT;
							break;
						case 6:
							//---match
							ch=matchchar;
							break;
					}//of bctr switch
	
	
					for(r=0;r<run;r++){
						fputc((UINT8)ch,out);
						outbytes++;
						//---update bst
						updatebst();
						//---get mtf position
						getmtf();
						//---move to front and update MTF context
						updatemtf();
						//---load ch
						ch=rank0;
					}//of run	
				}//of stopblock
			}//of blockmode=1
		}//of compressed or last block
	}//of stop

	//---close Files
	fclose(in);
	fclose(out);

	//---add header bytes
	inbytes+=headerlength;

	//---display result
	printf("%d Bytes decoded to %d Bytes\n",inbytes,outbytes);
	printf("Ratio: %f = %f bpb\n",(float)outbytes/inbytes,((float)outbytes/inbytes)*8);
	totaltime=((double)(clock()-begintime))/CLOCKS_PER_SEC;
	printf("Time : %.3f sec\n",totaltime);
	printf("Speed: %.3f Mbyte/s in, %.3f Mbyte/s out\n",(inbytes/totaltime)/(1024*1024),(outbytes/totaltime)/(1024*1024));
	//---end of decoding
	return 0;
}
//------------------------------------------------------------------------------
// USAGE
//------------------------------------------------------------------------------
void usage()
{
	printf("\nUsage: lcssr [options] <infile> <outfile>");
	printf("\n");
	printf("\nOptions are:");
	printf("\n-b[0-9] Buffer length mode (default=0), requires buffer length x 9 free memory.");
	printf("\n            mode / buffer lengths in MB are:");
	printf("\n            0=1, 1=2, 2=4, 3=8, 4=16, 5=32, 6=64, 7=128, 8=256, 9=455");
	printf("\n-l[0-9] Suffix match limit mode (default=7).");
	printf("\n            mode / match length limits in bytes are:");
	printf("\n            0=0, 1=1, 2=2, 3=4, 4=8, 5=16, 6=32, 7=64, 8=128, 9=256");
	printf("\n-d      Decodes <infile> to <outfile>");
	printf("\n");
	printf("\nEncoding example: lcssr -b3 -l6 myfile.ext myfile.ext.lcssr");
	printf("\nDecoding example: lcssr -d myfile.ext.lcssr myfile.ext");
	printf("\n");
	printf("\nThis program is free software; you can redistribute it and/or modify it under");
	printf("\nthe terms of the GNU General Public License. Input files with a size > 4 GB");
	printf("\nare not supported. Please report bugs to: frank.schwellinger@gmx.de\n");
}
//------------------------------------------------------------------------------
// MAIN
//------------------------------------------------------------------------------
SINT32 main(SINT32 argc,char *argv[] )
{
	SINT32 i;
	SINT32 codecmode,innum=0,outnum=0;
	FILE *in,*out;
	printf("lcssr v%d.%d - file (de-)compression using symbol ranking by longest common suffix\n",VERSION,SUBVERSION);
	//---default values
	codecmode=0;
	buffermode=0;
	limitmode=7;
	//---scan command line
	if(argc<3) {
		usage();
		exit(1);
	}
	for(i=1;i<argc;i++) {
		char *sz=argv[i];
		if(*sz=='-') {
			switch(*(sz+1)) {
				case 'b':
				case 'B':
					//---Buffer mode
					switch(*(sz+2)) {
						case '0':
							buffermode=0;
							break;
						case '1':
							buffermode=1;
							break;
						case '2':
							buffermode=2;
							break;
						case '3':
							buffermode=3;
							break;
						case '4':
							buffermode=4;
							break;
						case '5':
							buffermode=5;
							break;
						case '6':
							buffermode=6;
							break;
						case '7':
							buffermode=7;
							break;
						case '8':
							buffermode=8;
							break;
						case '9':
							buffermode=9;
							break;
						default:
							printf("bad argument: allowed is buffermode 0-9 \n");
							exit(1);
					}
					break;
				case 'l':
				case 'L':
					//---Limit mode
					switch(*(sz+2)) {
						case '0':
							limitmode=0;
							break;
						case '1':
							limitmode=1;
							break;
						case '2':
							limitmode=2;
							break;
						case '3':
							limitmode=3;
							break;
						case '4':
							limitmode=4;
							break;
						case '5':
							limitmode=5;
							break;
						case '6':
							limitmode=6;
							break;
						case '7':
							limitmode=7;
							break;
						case '8':
							limitmode=8;
							break;
						case '9':
							limitmode=9;
							break;
						default:
							printf("bad argument: allowed is limitmode 0-9\n");
							exit(1);
					}
					break;
				case 'd':
				case 'D':
					//---decode
					if(argc!=4) {
						usage();
						exit(1);
					} else {
						codecmode=1;
						break;
					}
				default:
					printf("unknown argument\n");
					exit(1);
			}
		} else {
			if(argv[i+2]!=NULL) {
				printf("bad arguments\n");
				exit(1);
			} else {
				innum=i;
				outnum=i+1;
				break;
			}
		}
	}
	//---test input file
	in=fopen(argv[innum],"rb");
	if(in==NULL) {
		printf("can't open input file\n");
		exit(1);
	}
	fclose(in);
	//---test output file
	out=fopen(argv[outnum],"wb");
	if(out==NULL) {
		printf("can't open output file\n");
		exit(1);
	}
	fclose(out);
	//---encode/decode
	if(codecmode==0)
		return encode(argv[innum],argv[outnum]);
	if(codecmode==1)
		return decode(argv[innum],argv[outnum]);
}
//------------------------------------------------------------------------------
// END OF PROGRAM
//------------------------------------------------------------------------------
